import { ResponseImpl, type Response } from './response'
import { type ApplicationContext, Dino, type EnvironmentConfiguration, Logger, ObjectHelper } from 'dino-core'
import { type RouterWithHandler } from './types'
import { type Request } from 'express'
import { FinalResponseMiddleware } from './middlewares/FinalResponseMiddleware'
import { AwsRequestAdaptor } from './request/AwsRequestAdaptor'

export class Server {
  private applicationContext!: ApplicationContext
  private environment!: EnvironmentConfiguration
  private intervalTimer!: NodeJS.Timeout
  private router!: RouterWithHandler
  private finalResponseHandler!: FinalResponseMiddleware
  private serverType!: string
  private cloudProvider!: string
  private awsRequestAdaptor!: AwsRequestAdaptor

  private async load(config?: string): Promise<void> {
    this.applicationContext = await Dino.run(config)
    this.environment = this.applicationContext.resolve('environment')
    this.awsRequestAdaptor = new AwsRequestAdaptor(this.environment)
    this.intervalTimer = setInterval(this.discoverRouter.bind(this), 10)
    this.finalResponseHandler = new FinalResponseMiddleware(this.environment)
    this.serverType = this.environment.getOrDefault('dino:server:type', 'express')
    this.cloudProvider = this.environment.getOrDefault('dino:cloud:provider', 'google')
  }

  public async run(req: Request, res: Response): Promise<any> {
    await this.load(undefined)
    if (this.serverType === 'express') {
      // the context will be loaded and the express server will be started
      await Promise.resolve()
    } else {
      return await this.handleServerlessMode(req, res)
    }
    // const exposedAs = environment.getOrDefault('dino:cloud:expose', 'api');
  }

  private async handleServerlessMode(req: Request, res: Response): Promise<any> {
    return await new Promise((resolve, reject): void => {
      let request: Request
      let response: Response
      if (this.serverType === 'serverless' && this.cloudProvider === 'aws') {
        // adapt aws request to express request
        request = this.awsRequestAdaptor.adapt(req)
        // eslint-disable-next-line @typescript-eslint/consistent-type-assertions
        response = new ResponseImpl(resolve, request)
      } else {
        request = req
        response = res
      }
      if (ObjectHelper.isDefined(this.router)) {
        Logger.debug('router has already been initialized...')
        this.router.handle(request, response, this.finalResponseHandler.handle(request, response, resolve, reject))
      } else {
        Logger.info('router not found initialize it...')
        const resolver = (): any => {
          this.router = this.applicationContext.resolve('router')
          clearInterval(this.intervalTimer)
          this.router.handle(request, response, this.finalResponseHandler.handle(request, response, resolve, reject))
        }
        setTimeout(resolver.bind(this), 500)
      }
    })
  }

  private discoverRouter(): void {
    try {
      this.router = this.applicationContext.resolve('router')
      if (ObjectHelper.isDefined(this.router)) {
        clearInterval(this.intervalTimer)
      }
    } catch (error) {
      // nothing to do
    }
  }
}
