//    Copyright 2018 Quirino Brizi [quirino.brizi@gmail.com]
//
//    Licensed under the Apache License, Version 2.0 (the "License");
//    you may not use this file except in compliance with the License.
//    You may obtain a copy of the License at
//
//        http://www.apache.org/licenses/LICENSE-2.0
//
//    Unless required by applicable law or agreed to in writing, software
//    distributed under the License is distributed on an "AS IS" BASIS,
//    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//    See the License for the specific language governing permissions and
//    limitations under the License.

import { type Response as EResponse } from 'express'
import { type GenericObject } from './types'
import { type HttpException } from './exception/http.exception'
import { type CookieOptions, type Locals, type Application, type Request, type ParamsDictionary } from 'express-serve-static-core'
import { type OutgoingHttpHeaders } from 'http'
import { type Socket } from 'net'
import { type ParsedQs } from 'qs'
import { ObjectHelper } from 'dino-core'

export type ResponsePayload = GenericObject | string | undefined
export type ErrorResponsePayload = GenericObject | HttpException | Error | undefined

// eslint-disable-next-line @typescript-eslint/no-extraneous-class
class EventEmitter {
  static captureRejectionSymbol: any
}

interface Chunk {
  chunk: any
  encoding: BufferEncoding
}

/**
 * The response that is returned from the API handler.
 * @public
 */
export interface Response extends EResponse {
  body: ResponsePayload
  // serviceErrorResponse: ErrorResponsePayload;
}

export class ResponseImpl implements Response {
  public body: ResponsePayload
  public statusCode!: number
  public statusMessage: string = ''
  public strictContentLength: boolean = false
  // eslint-disable-next-line @typescript-eslint/consistent-type-assertions
  public app: Application<Record<string, any>> = {} as Application
  public locals: Record<string, any> & Locals = {}
  public charset: string = 'UTF-8'

  public chunkedEncoding: boolean = false
  public shouldKeepAlive: boolean = false
  public useChunkedEncodingByDefault: boolean = false
  public sendDate: boolean = false
  public finished: boolean = false
  public connection: Socket | null = null
  public socket: Socket | null = null
  public writable: boolean = false
  public writableEnded: boolean = false
  public writableFinished: boolean = false
  public writableHighWaterMark: number = 0
  public writableLength: number = 0
  public writableObjectMode: boolean = false
  public writableCorked: number = 0
  public destroyed: boolean = false
  public closed: boolean = false
  public errored: Error | null = null
  public writableNeedDrain: boolean = false

  public headersSent: boolean = false
  public headers: GenericObject = {}
  public readonly req: Request<ParamsDictionary, any, any, ParsedQs, Record<string, any>>
  private readonly callback: (...args: any) => void

  constructor(callback: (...args: any) => void, req: Request<ParamsDictionary, any, any, ParsedQs, Record<string, any>>) {
    this.req = req
    this.callback = callback
  }

  sendfile(_path: unknown, _options?: unknown, _fn?: unknown): void {
    throw new Error('Method not implemented.')
  }

  [EventEmitter.captureRejectionSymbol]?(_error: Error, _event: string, ..._args: any[]): void {
    throw new Error('Method not implemented.')
  }

  public status(code: number): any {
    this.statusCode = code
    return this
  }

  sendStatus(_code: number): this {
    throw new Error('Method not implemented.')
  }

  links(_links: any): this {
    throw new Error('Method not implemented.')
  }

  public send(body?: GenericObject): any {
    this.callback(body)
    return this
  }

  public json(body?: GenericObject): any {
    return this.send(body)
  }

  public jsonp(body?: GenericObject): any {
    return this.send(body)
  }

  public sendFile(_path: unknown, _options?: unknown, _fn?: unknown): void {
    throw new Error('Method not implemented.')
  }

  download(_path: unknown, _filename?: unknown, _options?: unknown, _fn?: unknown): void {
    throw new Error('Method not implemented.')
  }

  contentType(_type: string): this {
    throw new Error('Method not implemented.')
  }

  type(_type: string): this {
    throw new Error('Method not implemented.')
  }

  public format(obj: GenericObject): this {
    const req = this.req
    const next = req.next

    // const keys = Object.keys(obj).filter(function (v) {
    //   return v !== 'default'
    // })

    // const key = keys.length > 0 ? req.accepts(keys) : undefined

    // this.vary('Accept')

    // if (key !== undefined) {
    //   this.set('Content-Type', normalizeType(key).value)
    //   obj[key](req, this, next)
    // } else
    if (obj.default !== undefined) {
      obj.default(req, this, next)
    }
    //  else {
    //   next(
    //     createError(406, {
    //       types: normalizeTypes(keys).map(function (o) {
    //         return o.value
    //       })
    //     })
    //   )
    // }

    return this
  }

  attachment(_filename?: string | undefined): this {
    throw new Error('Method not implemented.')
  }

  public set(field: string, value?: unknown): this {
    this.headers[field] = value
    return this
  }

  header(field: string, value?: unknown): this {
    if (ObjectHelper.isDefined(value)) {
      this.headers[field] = value
    }
    return this
  }

  get(field: string): string | undefined {
    return this.headers[field]
  }

  clearCookie(_name: string, _options?: CookieOptions | undefined): this {
    throw new Error('Method not implemented.')
  }

  cookie(_name: unknown, _val: unknown, _options?: unknown): this {
    throw new Error('Method not implemented.')
  }

  location(_url: string): this {
    throw new Error('Method not implemented.')
  }

  redirect(_url: unknown, _status?: unknown): void {
    throw new Error('Method not implemented.')
  }

  render(_view: unknown, _options?: unknown, _callback?: unknown): void {
    throw new Error('Method not implemented.')
  }

  vary(_field: string): this {
    throw new Error('Method not implemented.')
  }

  append(field: string, value?: string | string[] | undefined): this {
    this.set(field, value)
    return this
  }

  assignSocket(_socket: Socket): void {
    throw new Error('Method not implemented.')
  }

  detachSocket(_socket: Socket): void {
    throw new Error('Method not implemented.')
  }

  writeContinue(_callback?: (() => void) | undefined): void {
    throw new Error('Method not implemented.')
  }

  writeEarlyHints(_hints: Record<string, string | string[]>, _callback?: (() => void) | undefined): void {
    throw new Error('Method not implemented.')
  }

  writeHead(_statusCode: unknown, _statusMessage?: unknown, _headers?: unknown): this {
    throw new Error('Method not implemented.')
  }

  writeProcessing(): void {
    throw new Error('Method not implemented.')
  }

  setTimeout(_msecs: number, _callback?: (() => void) | undefined): this {
    throw new Error('Method not implemented.')
  }

  setHeader(name: string, value: string | number | readonly string[]): this {
    return this.set(name, value)
  }

  appendHeader(_name: string, _value: string | readonly string[]): this {
    throw new Error('Method not implemented.')
  }

  getHeader(name: string): string | number | string[] | undefined {
    return this.headers[name]
  }

  getHeaders(): OutgoingHttpHeaders {
    return this.headers
  }

  getHeaderNames(): string[] {
    throw new Error('Method not implemented.')
  }

  hasHeader(_name: string): boolean {
    throw new Error('Method not implemented.')
  }

  removeHeader(_name: string): void {
    throw new Error('Method not implemented.')
  }

  addTrailers(_headers: OutgoingHttpHeaders | ReadonlyArray<[string, string]>): void {
    throw new Error('Method not implemented.')
  }

  flushHeaders(): void {
    throw new Error('Method not implemented.')
  }

  _write(_chunk: any, _encoding: BufferEncoding, _callback: (error?: Error | null | undefined) => void): void {
    throw new Error('Method not implemented.')
  }

  _writev?(_chunks: Chunk[], _callback: (error?: Error | null | undefined) => void): void {
    throw new Error('Method not implemented.')
  }

  _construct?(_callback: (error?: Error | null | undefined) => void): void {
    throw new Error('Method not implemented.')
  }

  _destroy(_error: Error | null, _callback: (error?: Error | null | undefined) => void): void {
    throw new Error('Method not implemented.')
  }

  _final(_callback: (error?: Error | null | undefined) => void): void {
    throw new Error('Method not implemented.')
  }

  write(_chunk: unknown, _encoding?: unknown, _callback?: unknown): boolean {
    throw new Error('Method not implemented.')
  }

  setDefaultEncoding(_encoding: BufferEncoding): this {
    throw new Error('Method not implemented.')
  }

  end(_chunk?: unknown, _encoding?: unknown, _cb?: unknown): this {
    throw new Error('Method not implemented.')
  }

  cork(): void {
    throw new Error('Method not implemented.')
  }

  uncork(): void {
    throw new Error('Method not implemented.')
  }

  destroy(_error?: Error | undefined): this {
    throw new Error('Method not implemented.')
  }

  addListener(_event: unknown, _listener: unknown): this {
    throw new Error('Method not implemented.')
  }

  emit(_event: unknown, _src?: unknown, ..._rest: unknown[]): boolean {
    throw new Error('Method not implemented.')
  }

  on(_event: unknown, _listener: unknown): this {
    throw new Error('Method not implemented.')
  }

  once(_event: unknown, _listener: unknown): this {
    throw new Error('Method not implemented.')
  }

  prependListener(_event: unknown, _listener: unknown): this {
    throw new Error('Method not implemented.')
  }

  prependOnceListener(_event: unknown, _listener: unknown): this {
    throw new Error('Method not implemented.')
  }

  removeListener(_event: unknown, _listener: unknown): this {
    throw new Error('Method not implemented.')
  }

  pipe<T extends NodeJS.WritableStream>(_destination: T, _options?: { end?: boolean | undefined } | undefined): T {
    throw new Error('Method not implemented.')
  }

  compose<T extends NodeJS.ReadableStream>(
    _stream: ((source: any) => void) | T | Iterable<T> | AsyncIterable<T>,
    _options?: { signal: AbortSignal } | undefined
  ): T {
    throw new Error('Method not implemented.')
  }

  off(_eventName: string | symbol, _listener: (...args: any[]) => void): this {
    throw new Error('Method not implemented.')
  }

  removeAllListeners(_event?: string | symbol | undefined): this {
    throw new Error('Method not implemented.')
  }

  setMaxListeners(_n: number): this {
    throw new Error('Method not implemented.')
  }

  getMaxListeners(): number {
    throw new Error('Method not implemented.')
  }

  // eslint-disable-next-line @typescript-eslint/ban-types
  listeners(_eventName: string | symbol): Function[] {
    throw new Error('Method not implemented.')
  }

  // eslint-disable-next-line @typescript-eslint/ban-types
  rawListeners(_eventName: string | symbol): Function[] {
    throw new Error('Method not implemented.')
  }

  // eslint-disable-next-line @typescript-eslint/ban-types
  listenerCount(_eventName: string | symbol, _listener?: Function | undefined): number {
    throw new Error('Method not implemented.')
  }

  eventNames(): Array<string | symbol> {
    throw new Error('Method not implemented.')
  }
}
